/*
 * JSVG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */


package com.jsevy.jxml;

import org.w3c.dom.Attr;
import org.w3c.dom.DOMException;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.TypeInfo;

/**
 * Class representing an XML attribute associated with an XML node.
 * @author jsevy
 *
 */
public class XMLAttribute extends XMLNode
                            implements Attr
{
    
    protected Element ownerElement = null;
    protected boolean isId = false;
    protected boolean specified = false;
    
    
    
    /**
     * Create a new attribute with the specified name.
     * @param name      Name for the attribute
     * @param value     Value of the attribute
     */
    public XMLAttribute(String name)
    {
        super(name);
    }
    
    
    /**
     * Create a new attribute with the specified name and value.
     * @param name      Name for the attribute
     * @param value     Value of the attribute
     */
    public XMLAttribute(String name, String value)
    {
        super(name, value);
        
        // since value has been specified, specified = true
        specified = true;
    }
    
    
    /*
    @Override
    public Node cloneNode(boolean deep)
    {
        XMLAttribute newNode = new XMLAttribute(this.name, this.value.toString());
        
        // clone kids recursively - should be rare for attribute to have children... but spec says,
        // "Cloning an Attr always clones its children, since they represent its value, no matter 
        // whether this is a deep clone or not." Ok then... Still seems weird.
        for (int i = 0; i < children.getLength(); i++)
        {
            Node child = children.item(i);
            Node newChild = child.cloneNode(deep);
            newNode.appendChild(newChild);
        }
        
        return newNode;
    }
    */


    @Override
    public short getNodeType()
    {
        return Node.ATTRIBUTE_NODE;
    }



    @Override
    public String getName()
    {
        return this.getNodeName();
    }


    @Override
    public boolean getSpecified()
    {
        return specified;
    }


    @Override
    public String getValue()
    {
        return this.getNodeValue();
    }


    @Override
    public void setValue(String value) throws DOMException
    {
        this.setNodeValue(value);
    }
    
    
    @Override
    public void setNodeValue(String value) throws DOMException
    {
        super.setNodeValue(value);
        specified = true;
    }


    @Override
    public Element getOwnerElement()
    {
        return ownerElement;
    }
    
    
    protected void setOwnerElement(Element ownerElement)
    {
        this.ownerElement = ownerElement;
    }


    @Override
    public TypeInfo getSchemaTypeInfo()
    {
        // TODO Auto-generated method stub
        return null;
    }


    @Override
    public boolean isId()
    {
        return isId;
    }
    
    
    public void setId(boolean isId)
    {
        this.isId = isId;
    }
    
    
    @Override
    public String getTextContent() throws DOMException
    {
        // return depends on subclass, so overridden
        // for Attr node, return concatenation of the textContent attribute value of every child node,
        // excluding COMMENT_NODE and PROCESSING_INSTRUCTION_NODE nodes. This is the empty string if the 
        // node has no children. Note that this is unusual - generally an Attr won't have children, though it
        // can have entity references.... But weirdly the spec doesn't have an Attr return its value as its text
        // content in the "normal" case (no children)...
        
        StringBuffer text = new StringBuffer();
        
        for (int i = 0; i < this.getChildNodes().getLength(); i++)
        {
            // recursively get text from nodes except comment nodes and processing instruction nodes
            Node child = this.getChildNodes().item(i);
            short nodeType = child.getNodeType();
            if ((nodeType != Node.COMMENT_NODE) && (nodeType != Node.PROCESSING_INSTRUCTION_NODE))
            {
                String childText = child.getTextContent();
                
                // filter out null results since StringBuffer appends the string "null"... :-\
                if (childText != null)
                    text.append(childText);
            }
        }
        
        return text.toString();
    }
}
